@@ -1,7 +1,7 @@
 #!/usr/bin/perl
 package axmud;
 
-# Copyright (C) 2011-2018 A S Lewis
+# Copyright (C) 2011-2019 A S Lewis
 #
 # This program is free software: you can redistribute it and/or modify it under the terms of the GNU
 # General Public License as published by the Free Software Foundation, either version 3 of the
@@ -37,21 +37,21 @@ use vars qw(
     $AUTHORS $COPYRIGHT $URL $DESCRIP $NAME_FILE @COMPAT_FILE_LIST @COMPAT_DIR_LIST @COMPAT_EXT_LIST
     $BLIND_MODE_FLAG $SAFE_MODE_FLAG $TEST_MODE_FLAG @TEST_MODE_LOGIN_LIST $TEST_MODE_CMD_FLAG
     $TEST_TERM_MODE_FLAG $TEST_GLOB_MODE_FLAG $TEST_REGEX_FLAG $TEST_REGEX_ERROR
-    $TEST_PRE_CONFIG_FLAG $TEST_CTRL_SEQ_FLAG @LICENSE_LIST @CREDIT_LIST $TOP_DIR $SHARE_DIR
-    $DEFAULT_DATA_DIR $DATA_DIR $CLIENT
+    $TEST_PRE_CONFIG_FLAG $TEST_CTRL_SEQ_FLAG $TEST_MODEL_FLAG $TEST_MODEL_TIME $DEFAULT_ROOM
+    $DEFAULT_EXIT @LICENSE_LIST @CREDIT_LIST $TOP_DIR $SHARE_DIR $DEFAULT_DATA_DIR $DATA_DIR $CLIENT
 );
 
 $SCRIPT = 'Axmud';              # Name used in system messages
-$VERSION = '1.1.405';           # Version number for this client
-$DATE = '24 Dec 2018';
+$VERSION = '1.2.0';             # Version number for this client
+$DATE = '31 Mar 2019';
 $NAME_SHORT = 'axmud';          # Lower-case version of $SCRIPT; same as the package name above
 $NAME_ARTICLE = 'an Axmud';     # Name with an article
 $BASIC_NAME = 'Axbasic';        # Name of Axmud's built-in scripting library
 $BASIC_ARTICLE = 'an Axbasic';  # Name with an article
-$BASIC_VERSION = '1.001';       # Version number for the Axbasic library
+$BASIC_VERSION = '1.002';       # Version number for the Axbasic library
 $AUTHORS = 'A S Lewis';
-$COPYRIGHT = 'Copyright 2011-2018 A S Lewis';
-$URL = 'http://axmud.sourceforge.net/';
+$COPYRIGHT = 'Copyright 2011-2019 A S Lewis';
+$URL = 'http://axmud.sourceforge.io/';
 $DESCRIP = 'A modern MUD client for MS Windows, Linux and *BSD';
 
 # Name used in headers of Axmud config/data files
@@ -102,7 +102,20 @@ $TEST_PRE_CONFIG_FLAG = FALSE;
 # Simple telnet mode: If $TEST_CTRL_SEQ_FLAG is TRUE, VT100 control sequences (except colour/style
 #   sequences) are ignored (equivalent to GA::Client->useCtrlSeqFlag being FALSE)
 $TEST_CTRL_SEQ_FLAG = FALSE;
+# Automatic world model test mode: If $TEST_MODEL_FLAG is true, various parts of the code run a
+#   silent world model test from time to time, displaying output only if the test fails
+$TEST_MODEL_FLAG = FALSE;
+# The time of the last world model test (matches GA::Session->sessionTime)
+$TEST_MODEL_TIME = 0;
 
+# Default room object (GA::ModelObj::Room) and default exit object (GA::Obj::Exit) used to provide
+#   default values for IVs. (In order to reduce the size of the world model as much as possible,
+#   many IVs in room/exit objects don't actually exist in the blessed reference until they're given
+#   a non-default value)
+$DEFAULT_ROOM = undef;
+$DEFAULT_EXIT = undef;
+
+# Licence and credits
 @LICENSE_LIST = (
     'This program is free software; you can redistribute it and/or modify it under',
     'the terms of the GNU General Public License as published by the Free Software',
@@ -123,20 +136,21 @@ $TEST_CTRL_SEQ_FLAG = FALSE;
     'Chat task based on Kildclient plugin by Eduardo M Kalinowski',
     'Pathfinding algorithms based on AI::Pathfinding::AStar by Aaron Dalton',
     'Roman numeral conversion based on Text::Roman by Stanislaw Pusep',
-    'Simple list code copied (unmodified) from Gtk2::Ex::Simple::List by muppet, Ross',
-    '   McFarland and Gavin Brown',
+    'Simple list code based on Gtk3::SimpleList by Thierry Vignaud',
     'Telnet code based on Net::Telnet by Jay Rogers',
+    'Window manager control code based on X11::WMCtrl by Gavin Brown',
     'Images/icons by Dave Stokes, www.fatcow.com and A S Lewis. License information',
-    '   and full attributions can be found in /images/COPYING and /icons/COPYING',
+    '   and full attributions can be found in ../share/images/COPYING and',
+    '   ../share/icons/COPYING',
     'Sound by KevanGC, AirMan, AngryFlash, battlestar10, Brandondorf, Cam Martinez,',
     '   Christopher, Conor, Daniel Simon, DrumM8, G-rant, Grant Evans, Grandpa,',
     '   J Blow, J Bravo, Kevan, KevanGC, Lisa Redfern, Maximilien, Mike Koenig,',
     '   Muska666, Pool Shot, PsychoBird, RA The Sun God, Ragdoll485, Samantha Enrico,',
     '   Simon Craggs, Snore Man, Sonidor, Sound Explorer, Stephan, Sweeper, tamskp,',
     '   Tim Fryer, Vladimir, Willem Hunt and Yannick Lemieux. License information and',
-    '   full attributions can be found in /items/sounds/COPYING',
+    '   full attributions can be found in ../share/items/sounds/COPYING',
     'Documentation and help files by A S Lewis. Licence information can be found in',
-    '   /help/COPYING',
+    '   ../share/docs/COPYING and ../share/help/COPYING',
 );
 
 # External dependencies (Glib is commented out as it's already been used)
@@ -156,16 +170,8 @@ use File::Path qw(remove_tree);
 use File::ShareDir ':ALL';
 use File::ShareDir::Install;
 #use Glib qw(TRUE FALSE);
-use Glib::Object::Subclass;
-use Gnome2::Canvas;
-
-if ($^O ne 'MSWin32') {
-
-    # Wnck doesn't exist on MS Windows systems
-    eval "use Gnome2::Wnck";
-}
-
-use Gtk2 '-init';
+use GooCanvas2;
+use Gtk3 '-init';
 use HTTP::Tiny;
 use IO::Socket::INET;
 use IO::Socket::INET6;
@@ -245,7 +251,7 @@ $SIG{__DIE__} = sub {
             # Regex test initiated by GA::Client->regexCheck
             $TEST_REGEX_ERROR = $_[0];
 
-        } elsif (! $SAFE_MODE_FLAG) {
+        } else {
 
             # We don't know which GA::Session caused the Perl error, but we can leave Axmud in a
             #   (more or less) functional state by halting all client loops and session loops; the
@@ -267,22 +273,14 @@ $SIG{__DIE__} = sub {
 
 $SIG{__WARN__} = sub {
 
-    # v1.1.159 - filter out same warning messages caused by Perl module/Gtk+ issues, that would
-    #   otherwise spam the Client Console window, Session Console window and/or terminal
-    # As of this version, the 'Failed to parse menu bar accelerator' warning is displayed in a *BSD
-    #   terminal before being captured by $SIG{__WARN__}, so little we can do about it
-
     if ($TEST_REGEX_FLAG) {
 
         # Regex test initiated by GA::Client->regexCheck
         $TEST_REGEX_ERROR = $_[0];
 
     } elsif (
-        ! ($_[0] =~ m/gdk_pixbuf_from_pixdata\(\) called on/)
-        && ! ($_[0] =~ m/GdkPixbuf-LOG \*\*\:/)
-        && ! ($_[0] =~ m/Failed to parse menu bar accelerator/)
-        && ! ($_[0] =~ m/Argument.*isn\'t numeric in numeric ge.*Archive\/Extract\.pm line/)
-        && ! ($_[0] =~ m/g_object_u?n?ref\: assertion \'G_IS_OBJECT/)
+        # Warnings seen only on MS Windows
+        ! ($_[0] =~ m/attempt to override closure\-\>va_marshal/)
     ) {
         if ($CLIENT && ! $SAFE_MODE_FLAG) {
             $CLIENT->writePerlWarning(@_);
@@ -300,8 +298,8 @@ if (! $CLIENT || ! $CLIENT->start()) {
     exit 1;
 }
 
-# Start Gtk2's main loop
-Gtk2->main();
+# Start Gtk3's main loop
+Gtk3->main();
 
 END {
 
